/**
 ** Copyright (c) 2002,2009 LAAS/CNRS
 **
 ** Matthieu Herrb - Mon Apr 15 2002
 **
 **/

#include <math.h>
#include <stdio.h>

#include <rFlex.h>

#include "rflexConst.h"
#include "server/rflexHeader.h"
#include "odo.h"
#ifdef GYRO
#include "gyroLib.h"
#endif

#include "odoProbaAtrv.h"

#include "rflexClient.h"
#include "rflexCodels.h"

#include "h2mathLib.h"

/* Flag pour le debugage de l'odo 3D */
#define DEBUG3D 0
#define DEBUG 0

/***
 *** Call back functions for all relpy messages sent by rFlex
 ***/

/**
 ** Reply to screen dump request
 **/
int
sysLcdDumpRpl(struct reflex *reflex, int rv, unsigned char id, int ts,
	      unsigned char row, unsigned char nbytes,
	      unsigned char *data)
{
	static int count = 0;
	static FILE *fp = NULL;
	char fileName[256];

	int ii, i2, col;

	fprintf(stderr, "%s:%d:%s() - row:%d nbytes:%d count:%d\n",
		__FILE__, __LINE__, __FUNCTION__, row, nbytes, count);

	if (!fp) {
		sprintf(fileName, "reflexScreenDump-%04d.pgm", count++);
		fp = fopen(fileName, "w");

		if (!fp) {
			return(-1);
		}

		fprintf(fp, "P2 320 240 1\n");
	}

	col = 0;

	for (ii=0; ii<nbytes; ii++) {
		for (i2 = 0; i2<8; i2++) {
			if (col++ > 20) {
				fprintf(fp, "\n  ");
				col = 0;
			}
			fprintf(fp, " %d", (data[ii] & (1<<i2)) == 0);
		}
	}

	fprintf(fp, "\n");

	if (row==239) {
		fclose(fp);
		fp = NULL;
	}

	return(0);
}

/* ------------------------------------------------------------ */

/**
 ** Handle reply to system status (brakes + battery) request
 **/
int
sysStatusRpl(struct reflex *reflex, int rv, unsigned char id, int ts,
	     int volts, unsigned char brake)
{
	double batt;
#if 0
	fprintf(stderr, "SYS_STATUS: volts=%d brake=%d\n", volts, brake);
#endif
	/* Clear the active request flag */
	SDI_F->activeReq &= ~RFLEX_SYSTEM_GET_STATE;

	/* Handle brakes */
	if (SDI_F->systemState.brake != brake) {
		if (brake) {
			printf("Brakes ON\n");
			rflexStopTracking();
		} else {
			printf("Brakes OFF\n");
		}
		SDI_F->systemState.brake = brake;
	}
	/* Handle battery level
	   1.05 is a correction factor to correspond to the value on the rflex display */
	batt = 1.05 * volts/100.0;
	if (SDI_F->systemState.batteryLevel >= RFLEX_LOW_BATTERY_LEVEL &&
	    batt < RFLEX_LOW_BATTERY_LEVEL) {
		/* Battery level falled under acceptable minimum */
		printf("**** Low battery ***\n");
		/** XXX do something ? */
	}
	SDI_F->systemState.batteryLevel = batt;
	if (batt < RFLEX_LOW_BATTERY_LEVEL) {
	  SDI_F->systemState.batteryStatus = RFLEX_BATTERY_LOW;
	}
	else if (batt < RFLEX_MEDIUM_BATTERY_LEVEL) {
	  SDI_F->systemState.batteryStatus = RFLEX_BATTERY_MEDIUM;
	}
	else {
	  SDI_F->systemState.batteryStatus = RFLEX_BATTERY_HIGH;
	}

#ifdef B21R
	/* The maximum batteryLevel for robot B21R is 24 */
	SDI_F->systemState.highBatteryLevel = RFLEX_HIGH_BATTERY_LEVEL;
	SDI_F->systemState.mediumBatteryLevel = RFLEX_MEDIUM_BATTERY_LEVEL;
	SDI_F->systemState.lowBatteryLevel = RFLEX_LOW_BATTERY_LEVEL;
#endif
	return(0);
}

/* ------------------------------------------------------------ */

/**
 ** Handle unknown replies from system
 **/
int
sysUnknownRpl(struct reflex *reflex, unsigned char id,
	      unsigned char size, unsigned char *buffer,
	      unsigned char opcode)
{
	int ii;

	fprintf(stderr, "%s:%6d:%s() - id=%d opcode=%d : ",
		__FILE__, __LINE__, __FUNCTION__, id, opcode);

	for (ii=0; ii<size; ii++) {
		fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
	}
	fprintf(stderr, "\n");

	return(0);
}

/* ============================================================ */

/**
 ** Handle report from axis and compute odometry
 **/

int
motAxisSystemRpt (struct reflex *reflex,
		  unsigned char id, int ts,
		  unsigned char axis,
		  int pos, int vel, int acc, int trq)
{
	static int oldPos = -1, oldTheta = -1;
	static int haveV = 0, haveW = 0;
#define DEBUG 0
#if DEBUG>=1
	static int count;
#endif
	double dx, dtheta;
	double direction = 0.0;
	static double mpitch=0.0, oldPitch=-1;

#if 0
	fprintf(stderr, "%s:%6d:%s()\n", __FILE__, __LINE__, __FUNCTION__);
	fprintf(stderr, "\tid        = %d\n", id);
	fprintf(stderr, "\ttimeStamp = %d\n", ts);
	fprintf(stderr, "\taxis      = %d\n", axis);
	fprintf(stderr, "\tpos       = %d\n", pos);
	fprintf(stderr, "\tvel       = %d\n", vel);
	fprintf(stderr, "\tacc       = %d\n", acc);
	fprintf(stderr, "\ttrq       = %d\n", trq);
#endif

	/* This catches readAttitudeEstimation failing */
	if (SDI_F->odometryMode == RFLEX_ODO_3D)
	{
	    if (readAttitudeEstimation())
	    {
		printf("switching back to odo2D\n");
		SDI_F->odometryMode = RFLEX_ODO_2D;
	    }
	}
	else
	{
	    mpitch   = 0;
	    oldPitch = -1;
	    /* Prepare Storage of pitch and roll angles in SDI
	     * mpitch is used to compute the new position estimate, while
	     * attitude.pitch is needed to compute the next mpitch */
	    memset(&SDI_F->attitude, 0, sizeof(SDI_F->attitude));
	}

	SDI_F->odo.robot.pitch = SDI_F->attitude.pitch;
	/* NB: the following angle is the EULER roll angle */
	SDI_F->odo.robot.roll  = SDI_F->attitude.roll;

	/* Store raw data from rFlex in the SDI for debugging */
	SDI_F->rawAxis[axis].pos = pos;
	SDI_F->rawAxis[axis].vel = vel;
	SDI_F->rawAxis[axis].acc = acc;
	SDI_F->rawAxis[axis].trq = trq;

	switch (axis) {
	case 0:
		/* Read attitude estimation poster if ODO-3D asked */
		if (SDI_F->odometryMode == RFLEX_ODO_3D)
		{
		    if (oldPitch == -1)
			oldPitch = SDI_F->attitude.pitch;
		    else
		    {
			/* Mean pitch angle */
			mpitch = (oldPitch + SDI_F->attitude.pitch) / 2.0;
			oldPitch = SDI_F->attitude.pitch;
			/* Store pitch and roll angles in SDI
			 * mpitch is used to compute the new position estimate, while
			 * attitude.pitch is needed to compute the next mpitch*/
		    }
		}

		/* translation axis */
		SDI_F->odo.robot.v = STEP_TO_M(vel);/* speed */
		if (oldPos == -1) {
			/* first time */
			oldPos = pos;
		} else {
			/* radial motion */
			dx = STEP_TO_M(pos - oldPos);
			SDI_F->miscOdo.s += fabs(dx);

			/* Update global position */
			SDI_F->odo.robot.xRob += dx*cos(SDI_F->odo.robot.theta)
			    *cos(mpitch);
			SDI_F->odo.robot.yRob += dx*sin(SDI_F->odo.robot.theta)
			    *cos(mpitch);
			/* In bryant angles (which are used in Pom position
			 * structures), the pitch is negative if dz is positive
			 * */
			SDI_F->odo.robot.zRob += -dx*sin(mpitch);

#			if DEBUG3D
			printf("xC = %lf  ", SDI_F->odo.robot.xRob);
			printf("zC = %lf  ", SDI_F->odo.robot.zRob);
			printf("mpitch_Deg = %lf  ", mpitch*180.0/3.14159);
#			endif

			/* OLI : Maybe update x and y would be a thing to do */
			if (SDI_F->configRef.backFlag == GENPOS_FORWARD_MOTION)
			  direction = 1.0;
			else if (SDI_F->configRef.backFlag == GENPOS_BACKWARD_MOTION)
			  direction = -1.0;
			else direction = 0.0;

			SDI_F->odo.robot.xRef = SDI_F->odo.robot.xRob + direction*
					     RFLEX_DIST_TO_CONTROLED_POINT*cos(SDI_F->odo.robot.theta);
			SDI_F->odo.robot.yRef = SDI_F->odo.robot.yRob + direction*
					     RFLEX_DIST_TO_CONTROLED_POINT*sin(SDI_F->odo.robot.theta);

			/* store old axis value */
			oldPos = pos;
		}
		haveV = 1;
		break;


		/* rotation axis */
	case 1:
	        /* speed */
		SDI_F->odo.robot.w = STEP_TO_RAD(vel);

		/* first time */
		if (oldTheta == -1) {
		    oldTheta = pos;
		} /* read angle */
		else {
#ifdef GYRO
		  if (SDI_F->gyro.initDone &&
		      SDI_F->gyro.currentMode != RFLEX_GYRO_OFF) {
		    if (!(gyroReadCorrAngle(&SDI_F->gyro.gyroTheta))) {
		      SDI_F->gyro.currentMode = RFLEX_GYRO_OFF;
		      SDI_F->gyro.initDone = FALSE;
		    }
		  }
#endif
		  /* -- odo is reference */
		  if (SDI_F->gyro.currentMode == RFLEX_GYRO_OFF ||
		      (SDI_F->gyro.currentMode == RFLEX_GYRO_ON_IF_MOTION
		       && fabs(SDI_F->odo.robot.w) < 0.017
		       && fabs(SDI_F->odo.robot.v) < 0.01)) {

		    dtheta = STEP_TO_RAD(pos - oldTheta);
		    SDI_F->odo.robot.theta = angleLimit(SDI_F->odo.robot.theta + dtheta);
		    /* reset gyro offset to match odo */
		    SDI_F->gyro.gyroToRobotOffset = angleLimit(SDI_F->odo.robot.theta
							       - SDI_F->gyro.gyroTheta);
		    /* switch back to odo */
		    if (SDI_F->gyro.gyroOn) {
#if DEBUG>=1
		      printf ("-> odo (%5.2f d) v %5.2f w%5.3f\n",
			      RAD_TO_DEG(SDI_F->odo.robot.theta),
			      SDI_F->odo.robot.v,
			      SDI_F->odo.robot.w);
#endif
		      SDI_F->gyro.gyroOn = FALSE;
		    }
		  }


		  /* -- gyro is reference */
		  else {
		    SDI_F->odo.robot.theta = angleLimit(SDI_F->gyro.gyroTheta +
							SDI_F->gyro.gyroToRobotOffset);
		    /* switch back to gyro */
		    if (!SDI_F->gyro.gyroOn) {
#if DEBUG>=1
		      printf ("-> gyro (%5.2f d) v %5.2f w%5.3f\n",
			      RAD_TO_DEG(SDI_F->odo.robot.theta),
			      SDI_F->odo.robot.v,
			      SDI_F->odo.robot.w);
#endif
		      SDI_F->gyro.gyroOn = TRUE;
		    }
		  }

		  /*debug */
#if DEBUG>=1
		  if (count==10) {
		    printf ("MODE %s. gyro %5.3f + offset %5.3f = %5.3f =? %5.3f (old %d pos %d)\n",
			    SDI_F->gyro.gyroOn ? "GYRO" : "ODO",
			    RAD_TO_DEG(SDI_F->gyro.gyroTheta),
			    RAD_TO_DEG(SDI_F->gyro.gyroToRobotOffset),
			    RAD_TO_DEG(angleLimit(SDI_F->gyro.gyroTheta +
						  SDI_F->gyro.gyroToRobotOffset)),
			    RAD_TO_DEG(SDI_F->odo.robot.theta), oldTheta, pos);
		    count=0;
		  }
		  count++;
#endif

		  /* store old values */
		  oldTheta = pos;
		}
		haveW = 1;
		break;
	default:
		/* Bug */
		fprintf(stderr, "%s: unknown axis %d\n", __FUNCTION__, axis);
	}

	/* computes probabilistic odometry */
	/*---- This should be done in 3D too....----*/
	if (haveV && haveW) {
	  haveV = 0;
	  haveW = 0;
	  odoProba(&SDI_F->odo, RFLEX_AXIS_WIDTH, SDI_F->varParams.coeffLinAng,
		   EXEC_TASK_PERIOD(RFLEX_RFLEXTASK_NUM));
	}

	/* Store current timestamp */
	SDI_F->ts = ts;
	return(0);
}

/* ------------------------------------------------------------ */

int
motAxisGetSystemRpl (struct reflex *reflex, int rv,
		     unsigned char id, int ts,
		     unsigned char axis,
		     int pos, int vel, int acc, int trq)
{


	if (id == 255) {
		fprintf(stderr, "%s:%6d:%s()\n\tretval = %d\n",
			__FILE__, __LINE__, __FUNCTION__, rv);
		fprintf(stderr, "\tid        = %d\n", id);
		fprintf(stderr, "\ttimeStamp = %d\n", ts);
		fprintf(stderr, "\taxis      = %d\n", axis);
		fprintf(stderr, "\tpos       = %d\n", pos);
		fprintf(stderr, "\tvel       = %d\n", vel);
		fprintf(stderr, "\tacc       = %d\n", acc);
		fprintf(stderr, "\ttrq       = %d\n", trq);
	}

	return(0);
}

/* ------------------------------------------------------------ */

int
motAxisGetModelRpl (struct reflex *reflex, int rv,
		    unsigned char id, int ts,
		    unsigned char axis,
		    int pos, int vel, int acc, int trq)
{


	if (id == 255) {
		fprintf(stderr, "%s:%6d:%s()\n\tretval = %d\n",
			__FILE__, __LINE__, __FUNCTION__, rv);
		fprintf(stderr, "\tid        = %d\n", id);
		fprintf(stderr, "\ttimeStamp = %d\n", ts);
		fprintf(stderr, "\taxis      = %d\n", axis);
		fprintf(stderr, "\tpos       = %d\n", pos);
		fprintf(stderr, "\tvel       = %d\n", vel);
		fprintf(stderr, "\tacc       = %d\n", acc);
		fprintf(stderr, "\ttrq       = %d\n", trq);
	}


	return(0);
}

/* ------------------------------------------------------------ */

int
motAxisGetTargetRpl (struct reflex *reflex, int rv,
		     unsigned char id, int ts,
		     unsigned char axis,
		     int pos, int vel, int acc, int trq)
{


	if (id == 255) {
		fprintf(stderr, "%s:%6d:%s()\n\tretval = %d\n",
			__FILE__, __LINE__, __FUNCTION__, rv);
		fprintf(stderr, "\tid        = %d\n", id);
		fprintf(stderr, "\ttimeStamp = %d\n", ts);
		fprintf(stderr, "\taxis      = %d\n", axis);
		fprintf(stderr, "\tpos       = %d\n", pos);
		fprintf(stderr, "\tvel       = %d\n", vel);
		fprintf(stderr, "\tacc       = %d\n", acc);
		fprintf(stderr, "\ttrq       = %d\n", trq);
	}


	return(0);
}

/* ------------------------------------------------------------ */

/**
 ** Reply to SetDir request
 **/
int
motAxisSetDirRpl (struct reflex *reflex, int rv,
		  unsigned char id, int ts,
		  unsigned char axis)
{

	/* Clear active request flag */
	SDI_F->activeReq &= ~RFLEX_MOT_SET_DIR;

	return(0);
}

/* ------------------------------------------------------------ */

/**
 ** Reply to SetDefaults request
 **/
int
motSetDefaultsRpl (struct reflex *reflex, int rv,
		   unsigned char id, int ts)
{
	/* Clear active request flag */
	SDI_F->activeReq &= ~RFLEX_MOT_SET_DEFAULT;
	return(0);
}

/* ------------------------------------------------------------ */

/**
 ** Reply to Set Brake request
 **/
int
motBrakeSetRpl(struct reflex *reflex, int rv,
	       unsigned char id, int ts)
{
	/* Clear active reflex flag */
	SDI_F->activeReq &= ~RFLEX_MOT_SET_BRAKES;

	return(0);
}

/* ------------------------------------------------------------ */

/**
 ** Repley to Release Brake request
 **/
int
motBrakeReleaseRpl(struct reflex *reflex, int rv,
		   unsigned char id, int ts)
{
	/* Clear active reflex flag */
	SDI_F->activeReq &= ~RFLEX_MOT_REL_BRAKES;


	return(0);
}

/* ------------------------------------------------------------ */

int
motSetSystemRptRpl (struct reflex *reflex, int rv,
		    unsigned char id, int ts)
{
#if 0
	fprintf(stderr, "%s:%6d:%s()\n", __FILE__, __LINE__, __FUNCTION__);
#endif
	return(0);
}

/* ------------------------------------------------------------ */

/**
 ** Handle unknown motion subsystem replies
 **/
int
motUnknownRpl(struct reflex *reflex, unsigned char id,
	      unsigned char size,
	      unsigned char *buffer, unsigned char opcode)
{
	int ii;

	fprintf(stderr, "%s:%6d:%s() - id=%d opcode=%d : ",
		__FILE__, __LINE__, __FUNCTION__, id, opcode);

	for (ii=0; ii<size; ii++) {
		fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
	}
	fprintf(stderr, "\n");

	return(0);
}

/* ============================================================ */
/* ============================================================ */

/**
 ** Joystick get state request
 **/
int
jstkGetStateRpl(struct reflex *reflex, int rv, unsigned char id, int ts,
		int x, int y, unsigned char buttons)
{

#if DEBUG>=1
	fprintf(stderr, "JSTK_GET_STATE:\n");
	fprintf(stderr, "\tid        = %d\n", id);
	fprintf(stderr, "\ttimeStamp = %d\n", ts);
	fprintf(stderr, "\tbuttons   = %d\n", buttons);
#endif

	/* Clear active request flag */
	SDI_F->activeReq &= ~RFLEX_JOY_GET_STATE;

	SDI_F->joystickState.x = x;
	SDI_F->joystickState.y = y;
	SDI_F->joystickState.buttons = buttons;

	/* Force manual mode if needed, aborts tracking */
#ifdef FORCE_MANUAL_MODE
	if (buttons == RFLEX_JOYSTICK_BUTTON2 &&
	    SDI_F->systemState.mode != MANUAL_MODE) {

	        printf ("RFLEX : jstkGetStateRpl: joystick butt2 : switch to manual mode\n");
		SDI_F->systemState.mode = MANUAL_MODE;

		rflexStopTracking();

	}
#endif
	return(0);
}

/* ------------------------------------------------------------ */

/**
 ** Unknown joystick replies
 **/
int
jstkUnknownRpl(struct reflex *reflex, unsigned char id, unsigned char size,
	       unsigned char *buffer, unsigned char opcode)
{
	int ii;

	fprintf(stderr, "%s:%6d:%s() - id=%d opcode=%d : ",
		__FILE__, __LINE__, __FUNCTION__, id, opcode);

	for (ii=0; ii<size; ii++) {
		fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
	}
	fprintf(stderr, "\n");

	return(0);
}

/* ============================================================ */
/* ============================================================ */

/**
 ** Sonar report frequency request
 **/
int
sonarRptRpl(struct reflex *reflex, int rv, unsigned char id, int ts)
{
	/* clear active request flag */
	SDI_F->activeReq &= ~RFLEX_SONAR_REPORT;
	return(0);
}

/* ------------------------------------------------------------ */

/**
 ** Sonar update messages
 **/
int
sonarGetUpdateRpl (struct reflex *reflex, int rv, unsigned char id, int ts,
		   int count, unsigned char *sonarId,
		   unsigned short *sonarValue)
{
	int ii;

	fprintf(stderr, "SONAR_GET_UPDATE: count=%d timeStamp = %d\n",
		count, ts);

	for (ii=0; ii<count; ii++) {
		fprintf(stderr, "   SONAR %02X: %5d\n",
			sonarId[ii], sonarValue[ii]);
	}

	fprintf(stderr, "\n");
	return(0);
}

/* ------------------------------------------------------------ */

/**
 ** Sonar reports
 **/
int
sonarRpt (struct reflex *reflex, unsigned char id, int ts,
	  int count, unsigned char *sonarId, unsigned short *sonarValue)
{
	int ii, i, j;

	/* Update sonar data */
	for (ii=0; ii<count; ii++) {
#ifdef B21R
	  int index;
	  /* Extract index from sonar Id */
	  i = (sonarId[ii] & 0xf0) >> 4;
	  j = sonarId[ii] & 0x0f;

       	  if(i<6) {
	  /* from sonar 0 to sonar 23 : enclosure */
	  /* numerotation : sonarId/sonar number
	     door A : 00/0  01/1  02/2  03/3
             door B : 10/4  11/5  12/6  13/7
	     door C : 20/8  21/9  22/10 23/11
             door D : 30/12 31/13 32/14 33/15
	     door E : 40/16 41/17 42/18 43/19
             door F : 50/20 51/21 52/22 53/23
	   */
	  index=4*i+j;
	  } else {
	  /* from sonar 24 to sonar 47 : base */
	  /* numerotation : sonarId/sonar number
	     door a : 60/24 61/25 62/26
             door b : 70/27 71/28 72/29
	     door c : 80/30 81/31 82/32
             door d : 90/33 91/34 92/35
	     door e : A0/36 A1/37 A2/38
             door f : B0/39 B1/40 B2/41
	     door g : C0/42 C1/43 C2/44
             door h : D0/45 D1/46 D2/47
	   */
	  index = 3*i+j+6;
	  }
	  SDI_F->sonarRange.tab[index] = SONAR_TO_M(sonarValue[ii]);
#endif
#if defined(ATRV) || defined(ATRV_JUNIOR)
	  /* Extract index from sonar Id */
	  i = (sonarId[ii] & 0xf0) >> 4;
	  j = sonarId[ii] & 0x0f;
	  SDI_F->sonarRange.tab[4*i+j] = SONAR_TO_M(sonarValue[ii]);
#endif
#ifdef ATRV_JUNIOR
	  int index;
	  /* Extract index from sonar Id */
	  i = (sonarId[ii] & 0xf0) >> 4;
	  j = sonarId[ii] & 0x0f;
	  index = 6*i +j;
	  if (index > 10) index--;
	  SDI_F->sonarRange.tab[index] = SONAR_TO_M(sonarValue[ii]);
#endif
	}

	/* Update timestamp */
	SDI_F->ts = ts;
  	return(0);
}


/**
 ** Unknown sonar messages
 **/
int
sonarUnknownRpl(struct reflex *reflex, unsigned char id, unsigned char size,
		unsigned char *buffer, unsigned char opcode)
{
	int ii;

	fprintf(stderr, "%s:%6d:%s() - id=%d opcode=%d : ",
		__FILE__, __LINE__, __FUNCTION__, id, opcode);
	for (ii=0; ii<size; ii++) {
		fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
	}
	fprintf(stderr, "\n");

	return(0);
}

/*----------------------------------------------------------------------*/


/* ============================================================ */
/* ============================================================ */

#ifdef B21R

int
dioUpdateRpl (struct reflex *reflex, int rv, unsigned char id, int ts,
		   int count)
{

  fprintf(stderr, "%s() - rv=%d count=%d timeStamp = %d\n",
	  __FUNCTION__, rv, count, ts);

  if (ts<SDI_F->dioTS) {
    SDI_F->dioTS = 0; /* time stamp wrapped or reset, next req will get update */
  }
  else if (SDI_F->dioCount == count) {
    SDI_F->dioTS = ts;
  }

  return(0);
}

/* ------------------------------------------------------------ */

int
dioUpdate (struct reflex *reflex, unsigned char id, int ts,
	   unsigned char netId, unsigned char offset, unsigned char state)
{

  fprintf(stderr, "%s:%6d:%s() - netId:0x%02X offset:%d state:0x%02X\n",
	  __FILE__, __LINE__, __FUNCTION__,
	  netId, offset, state);

  SDI_F->dioCount++;

  return(0);
}

/* ------------------------------------------------------------ */

 int
dioRptRpl(struct reflex *reflex, int rv, unsigned char id, int ts)
{
  fprintf(stderr, "%s() - retval = %d\n", __FUNCTION__, rv);
 /*  SDI_F->activeReq &= ~RFLEX_IO_REPORT;  */
  return(0);
}

/* ------------------------------------------------------------ */

int
dioRpt (struct reflex *reflex, unsigned char id, int ts,
	unsigned char netId, unsigned char offset, unsigned char state)
{
  int door_num;
  int digit;

  /* some contactors could be active at the same times
   * in that case, state is the sum of the state
   * ex : 0x01 -> 0001 say that switch 0 is active
   *      0x0c -> 1100 say that swicth 2 and 3 are active
   */

  /*      fprintf(stderr, "%s:%6d:%s() - netId:0x%02X offset:%d state:0x %02X \n", */
  /*     	  __FILE__, __LINE__, __FUNCTION__, */
  /*     	  netId, offset, state); */

  /* need to check if the first digit before the door number is 4 (cf doc B21R 3-20)
     cause rflex send other message, at least at the init, except for 0x31,
     no information is available about these messages */

  digit = (netId & 0xf0) >> 4;


  /* digit == 4 : door contactor message */
  if(digit == 4) {
    door_num = netId & 0x0f;

/*     printf ("door %d state = %0x\n", door_num, (state & 0x0f)); */

    if((state & 0x0f)== 0x00) {

      if (SDI_F->doorSwitchs.tab[door_num][0])
	printf ("rflex door %d OK\n", door_num);

      SDI_F->doorSwitchs.tab[door_num][0] = 0;
      SDI_F->doorSwitchs.tab[door_num][1] = 0;
      SDI_F->doorSwitchs.tab[door_num][2] = 0;
      SDI_F->doorSwitchs.tab[door_num][3] = 0;

      SDI_F->systemState.shock = 0;
    }

    else {

      if((state & 0x01) == 0x01) {
	if (!SDI_F->doorSwitchs.tab[door_num][0])
	  printf ("rflex hurts door %d switch up right\n", door_num);
	SDI_F->doorSwitchs.tab[door_num][0] = 1;
/* 	SDI_F->systemState.shock = 1; */
      }

      if((state & 0x02) == 0x02) {
	if (!SDI_F->doorSwitchs.tab[door_num][0])
	  printf ("rflex hurts door %d switch up left\n", door_num);
	SDI_F->doorSwitchs.tab[door_num][1] = 2;
/* 	SDI_F->systemState.shock = 1; */
      }
      if((state & 0x04) == 0x04) {
	if (!SDI_F->doorSwitchs.tab[door_num][0])
	  printf ("rflex hurts door %d switch down right\n", door_num);
	SDI_F->doorSwitchs.tab[door_num][2] = 3;
/* 	SDI_F->systemState.shock = 1; */
      }
      if((state & 0x08) == 0x08) {
	if (!SDI_F->doorSwitchs.tab[door_num][0])
	  printf ("rflex hurts door %d switch down left\n", door_num);
	SDI_F->doorSwitchs.tab[door_num][3] = 4;
/* 	SDI_F->systemState.shock = 1; */
      }
    }
  }

  return(0);
}

/* ------------------------------------------------------------ */

int
dioUnknownRpl(struct reflex *reflex, unsigned char id, unsigned char size,
	      unsigned char *buffer, unsigned char opcode)
{
  int ii;

  fprintf(stderr, "%s:%6d:%s() - id=%d opcode=%d : ",
	  __FILE__, __LINE__, __FUNCTION__, id, opcode);
  for (ii=0; ii<size; ii++) {
    fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
  }
  fprintf(stderr, "\n");

  return(0);
}

/* ------------------------------------------------------------ */

#endif /* B21R */

/**
 ** handler for all other data coming from rFlex controler
 **/
int
defaultHandlePort(struct reflex *reflex, int size, unsigned char *buffer,
		  unsigned char port, unsigned char id, unsigned char opcode)
{
	int fd;
	struct pktPort *pkt;

	if (opcode==1) {
		fflush(stdout);
		fprintf(stderr, "[%s:%6d:%s() - size = %d]\n",
			__FILE__, __LINE__, __FUNCTION__, size);
		return(0);
	}
	else if (opcode==2) {
		fflush(stdout);
		fprintf(stderr, "[%s:%6d:%s() - unknown port = %d]\n",
			__FILE__, __LINE__, __FUNCTION__, port);
		return(0);
	}

	pkt = reflex->pktPort;

	fd = pktGetFd(pkt);

	if ((size>0) && (opcode==0)) {
		buffer[size] = 0;
		fprintf(stderr, "%s", buffer);
	}
	else if ((size>0) && (opcode==1)) {
		int ii;

		fprintf(stderr, "%s:%6d:%s() - unknown port=%d :",
			__FILE__, __LINE__, __FUNCTION__, id);
		for (ii=0; ii<size; ii++) {
			fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
		}
		fprintf(stderr, "\n");
		return(0);
	}
	return(0);
}

